#include "../../../include/GPU/GPU.h"
#include "SimpleHelper.hpp"
#include <limits.h>
#include <stdio.h>

namespace FileProcessor {
    using GPU::PositionU16;
    using GPU::SizeU16;

    struct SimpleTIMSize : private SimpleTIM {
        constexpr SimpleTIMSize() {
        }

        constexpr uint16_t getTextureWidth() const {
            return SimpleTIM::getTextureX();
        }

        constexpr uint16_t getTextureHeight() const {
            return SimpleTIM::getTextureY();
        }

        constexpr uint16_t getClutWidth() const {
            return SimpleTIM::getClutX();
        }

        constexpr uint16_t getClutHeight() const {
            return SimpleTIM::getClutY();
        }
    };

    struct SimpleTIMState {
        SimpleTIM     dst_info;
        SimpleTIMSize size_info;
        size_t        words_left; //32bit values

        constexpr SimpleTIMState(const SimpleTIM& dst_info) : dst_info(dst_info) {
        }
    };

    static void set_gpu_receive(const uint32_t* src, uint16_t x, uint16_t y, uint16_t w, uint16_t h) {
        GPU::DMA::Receive::prepare();
        GPU::DMA::Receive::set_dst(PositionU16(x, y), SizeU16(w, h)); 
        GPU::DMA::Receive::set_src(reinterpret_cast<const uintptr_t>(src));
    }

    static void set_gpu_receive_data(const uint32_t* src, SimpleTIMState& state) {
        const auto width  = state.size_info.getTextureWidth();
        const auto height = state.size_info.getTextureHeight();

        state.words_left = (width*height)/2;
        set_gpu_receive(src, state.dst_info.getTextureX(), state.dst_info.getTextureY(), width, height);
    }

    static bool parse_data(State::Configuration& config, SimpleTIMState& state, bool is_last) {
        auto block_count = (state.words_left >> 4);
        
        state.words_left &= 0b1111;
        while(block_count > 0) {
            const auto block_send = (block_count > UI16_MAX) ? UI16_MAX : block_count;

            // Send data!
            GPU::DMA::wait();
            GPU::DMA::Receive::start(block_send);
            block_count -= block_send;
        }

        if(is_last) {
            // Send words
            if(state.words_left > 0) {
                GPU::DMA::wait();
                GPU::DMA::Receive::start(1, state.words_left);
            }

            GPU::DMA::wait();
            GPU::DMA::end();
            return false;
        }

        return true;
    }

    static bool parse_header(State::Configuration& config, SimpleTIMState& state, bool is_last) {
        if(config.data_size >= sizeof(SimpleTIMSize)) {
            Helper::simple_read(state.size_info, config);
            
            //Check if we have a clut to care about
            if(state.size_info.getClutWidth() > 0) {
                //CLUTs are 16bit full color anyway
                printf("We found a CLUT which is not yet implemented!\n");
                return false;
            }

            //We have direct data
            else {
                set_gpu_receive_data(config.data_adr, state);
                return Helper::exchange_and_execute_process_function(parse_data, config, state, is_last);
            }
        }

        return !is_last;
    }

    State create(const uint32_t* data_adr, const SimpleTIM& file) {        
        return State::from(SimpleTIMState(file), data_adr, parse_header);
    }
}

#undef private